@@ -127,6 +127,12 @@ wireToTouchOwnershipEvent(xXITouchOwnershipEvent *in,
 static int
 wireToBarrierEvent(xXIBarrierEvent *in,
                    XGenericEventCookie *cookie);
+static int
+wireToPinchEvent(xXIGesturePinchEvent *in,
+                 XGenericEventCookie *cookie);
+static int
+wireToSwipeEvent(xXIGestureSwipeEvent *in,
+                 XGenericEventCookie *cookie);
 
 static /* const */ XEvent emptyevent;
 
@@ -253,7 +259,9 @@ static XExtensionVersion versions[] = { {XI_Absent, 0, 0},
  XI_Add_DeviceProperties_Minor},
 {XI_Present, 2, 0},
 {XI_Present, 2, 1},
-{XI_Present, 2, 2}
+{XI_Present, 2, 2},
+{XI_Present, 2, 3},
+{XI_Present, 2, 4},
 };
 
 /***********************************************************************
@@ -380,7 +388,7 @@ _XiCheckExtInit(
 	    return (-1);
 	}
 	((XInputData *) info->data)->vers =
-	    _XiGetExtensionVersion(dpy, "XInputExtension", info);
+	    _XiGetExtensionVersionRequest(dpy, "XInputExtension", info->codes->major_opcode);
     }
 
     if (_XiCheckVersion(info, version_index) < 0) {
@@ -1035,6 +1043,28 @@ XInputWireToCookie(
                 break;
             }
             return ENQUEUE_EVENT;
+        case XI_GesturePinchBegin:
+        case XI_GesturePinchUpdate:
+        case XI_GesturePinchEnd:
+            *cookie = *(XGenericEventCookie*)save;
+            if (!wireToPinchEvent((xXIGesturePinchEvent*)event, cookie))
+            {
+                printf("XInputWireToCookie: CONVERSION FAILURE!  evtype=%d\n",
+                        ge->evtype);
+                break;
+            }
+            return ENQUEUE_EVENT;
+        case XI_GestureSwipeBegin:
+        case XI_GestureSwipeUpdate:
+        case XI_GestureSwipeEnd:
+            *cookie = *(XGenericEventCookie*)save;
+            if (!wireToSwipeEvent((xXIGestureSwipeEvent*)event, cookie))
+            {
+                printf("XInputWireToCookie: CONVERSION FAILURE!  evtype=%d\n",
+                        ge->evtype);
+                break;
+            }
+            return ENQUEUE_EVENT;
         default:
             printf("XInputWireToCookie: Unknown generic event. type %d\n", ge->evtype);
 
@@ -1065,7 +1095,7 @@ sizeDeviceEvent(int buttons_len, int valuators_len,
 /* Return the size with added padding so next element would be
    double-aligned unless the architecture is known to allow unaligned
    data accesses.  Not doing this can cause a bus error on
-   MIPS N32. */
+   MIPS N32. */
 static int
 pad_to_double(int size)
 {
@@ -1143,6 +1173,9 @@ sizeDeviceClassType(int type, int num_elements)
         case XITouchClass:
             l = pad_to_double(sizeof(XITouchClassInfo));
             break;
+        case XIGestureClass:
+            l = pad_to_double(sizeof(XIGestureClassInfo));
+            break;
         default:
             printf("sizeDeviceClassType: unknown type %d\n", type);
             break;
@@ -1432,6 +1465,41 @@ copyBarrierEvent(XGenericEventCookie *in_cookie,
     return True;
 }
 
+
+static Bool
+copyGesturePinchEvent(XGenericEventCookie *cookie_in,
+                      XGenericEventCookie *cookie_out)
+{
+    XIGesturePinchEvent *in, *out;
+
+    in = cookie_in->data;
+
+    out = cookie_out->data = malloc(sizeof(XIGesturePinchEvent));
+    if (!out)
+        return False;
+
+    *out = *in;
+
+    return True;
+}
+
+static Bool
+copyGestureSwipeEvent(XGenericEventCookie *cookie_in,
+                      XGenericEventCookie *cookie_out)
+{
+    XIGestureSwipeEvent *in, *out;
+
+    in = cookie_in->data;
+
+    out = cookie_out->data = malloc(sizeof(XIGestureSwipeEvent));
+    if (!out)
+        return False;
+
+    *out = *in;
+
+    return True;
+}
+
 static Bool
 XInputCopyCookie(Display *dpy, XGenericEventCookie *in, XGenericEventCookie *out)
 {
@@ -1484,12 +1552,25 @@ XInputCopyCookie(Display *dpy, XGenericEventCookie *in, XGenericEventCookie *out
         case XI_RawButtonPress:
         case XI_RawButtonRelease:
         case XI_RawMotion:
+        case XI_RawTouchBegin:
+        case XI_RawTouchUpdate:
+        case XI_RawTouchEnd:
             ret = copyRawEvent(in, out);
             break;
         case XI_BarrierHit:
         case XI_BarrierLeave:
             ret = copyBarrierEvent(in, out);
             break;
+        case XI_GesturePinchBegin:
+        case XI_GesturePinchUpdate:
+        case XI_GesturePinchEnd:
+            ret = copyGesturePinchEvent(in, out);
+            break;
+        case XI_GestureSwipeBegin:
+        case XI_GestureSwipeUpdate:
+        case XI_GestureSwipeEnd:
+            ret = copyGestureSwipeEvent(in, out);
+            break;
         default:
             printf("XInputCopyCookie: unknown evtype %d\n", in->evtype);
             ret = False;
@@ -1518,6 +1599,7 @@ wireToDeviceEvent(xXIDeviceEvent *in, XGenericEventCookie* cookie)
     out = next_block(&ptr_lib, sizeof(XIDeviceEvent));
     out->display = cookie->display;
     out->type = in->type;
+    out->serial = cookie->serial;
     out->extension = in->extension;
     out->evtype = in->evtype;
     out->send_event = ((in->type & 0x80) != 0);
@@ -1603,6 +1685,9 @@ size_classes(xXIAnyInfo* from, int nclasses)
             case XITouchClass:
                 l = sizeDeviceClassType(XITouchClass, 0);
                 break;
+            case XIGestureClass:
+                l = sizeDeviceClassType(XIGestureClass, 0);
+                break;
         }
 
         len += l;
@@ -1612,6 +1697,8 @@ size_classes(xXIAnyInfo* from, int nclasses)
     return len;
 }
 
+#define FP3232_TO_DOUBLE(x) ((double) (x).integral + (double) (x).frac / (1ULL << 32))
+
 /* Copy classes from any into to->classes and return the number of bytes
  * copied. Memory layout of to->classes is
  * [clsptr][clsptr][clsptr][classinfo][classinfo]...
@@ -1720,10 +1807,9 @@ copy_classes(XIDeviceInfo* to, xXIAnyInfo* from, int *nclasses)
                     cls_lib->number = cls_wire->number;
                     cls_lib->label  = cls_wire->label;
                     cls_lib->resolution = cls_wire->resolution;
-                    cls_lib->min        = cls_wire->min.integral;
-                    cls_lib->max        = cls_wire->max.integral;
-                    cls_lib->value      = cls_wire->value.integral;
-                    /* FIXME: fractional parts */
+                    cls_lib->min        = FP3232_TO_DOUBLE(cls_wire->min);
+                    cls_lib->max        = FP3232_TO_DOUBLE(cls_wire->max);
+                    cls_lib->value      = FP3232_TO_DOUBLE(cls_wire->value);
                     cls_lib->mode       = cls_wire->mode;
 
                     to->classes[cls_idx++] = any_lib;
@@ -1744,8 +1830,7 @@ copy_classes(XIDeviceInfo* to, xXIAnyInfo* from, int *nclasses)
                     cls_lib->number     = cls_wire->number;
                     cls_lib->scroll_type= cls_wire->scroll_type;
                     cls_lib->flags      = cls_wire->flags;
-                    cls_lib->increment  = cls_wire->increment.integral;
-                    cls_lib->increment += (unsigned int)cls_wire->increment.frac/(double)(1ULL << 32);
+                    cls_lib->increment  = FP3232_TO_DOUBLE(cls_wire->increment);
 
                     to->classes[cls_idx++] = any_lib;
                 }
@@ -1766,6 +1851,21 @@ copy_classes(XIDeviceInfo* to, xXIAnyInfo* from, int *nclasses)
                     to->classes[cls_idx++] = any_lib;
                 }
                 break;
+            case XIGestureClass:
+                {
+                    XIGestureClassInfo *cls_lib;
+                    xXIGestureInfo *cls_wire;
+
+                    cls_wire = (xXIGestureInfo*)any_wire;
+                    cls_lib = next_block(&ptr_lib, sizeof(XIGestureClassInfo));
+
+                    cls_lib->type = cls_wire->type;
+                    cls_lib->sourceid = cls_wire->sourceid;
+                    cls_lib->num_touches = cls_wire->num_touches;
+
+                    to->classes[cls_idx++] = any_lib;
+                }
+                break;
         }
         len += any_wire->length * 4;
         ptr_wire += any_wire->length * 4;
@@ -1790,6 +1890,7 @@ wireToDeviceChangedEvent(xXIDeviceChangedEvent *in, XGenericEventCookie *cookie)
     cookie->data = out = malloc(sizeof(XIDeviceChangedEvent) + len);
 
     out->type = in->type;
+    out->serial = cookie->serial;
     out->display = cookie->display;
     out->extension = in->extension;
     out->evtype = in->evtype;
@@ -1822,6 +1923,7 @@ wireToHierarchyChangedEvent(xXIHierarchyEvent *in, XGenericEventCookie *cookie)
     out->info           = (XIHierarchyInfo*)&out[1];
     out->display        = cookie->display;
     out->type           = in->type;
+    out->serial         = cookie->serial;
     out->extension      = in->extension;
     out->evtype         = in->evtype;
     out->send_event = ((in->type & 0x80) != 0);
@@ -1862,6 +1964,7 @@ wireToRawEvent(XExtDisplayInfo *info, xXIRawEvent *in, XGenericEventCookie *cook
 
     out = next_block(&ptr, sizeof(XIRawEvent));
     out->type           = in->type;
+    out->serial         = cookie->serial;
     out->display        = cookie->display;
     out->extension      = in->extension;
     out->evtype         = in->evtype;
@@ -1912,6 +2015,7 @@ wireToEnterLeave(xXIEnterEvent *in, XGenericEventCookie *cookie)
     out->buttons.mask = (unsigned char*)&out[1];
 
     out->type           = in->type;
+    out->serial         = cookie->serial;
     out->display        = cookie->display;
     out->extension      = in->extension;
     out->evtype         = in->evtype;
@@ -1954,6 +2058,7 @@ wireToPropertyEvent(xXIPropertyEvent *in, XGenericEventCookie *cookie)
     cookie->data = out;
 
     out->type           = in->type;
+    out->serial         = cookie->serial;
     out->extension      = in->extension;
     out->evtype         = in->evtype;
     out->send_event = ((in->type & 0x80) != 0);
@@ -1974,6 +2079,7 @@ wireToTouchOwnershipEvent(xXITouchOwnershipEvent *in,
     cookie->data = out;
 
     out->type           = in->type;
+    out->serial         = cookie->serial;
     out->display        = cookie->display;
     out->extension      = in->extension;
     out->evtype         = in->evtype;
@@ -1990,8 +2096,6 @@ wireToTouchOwnershipEvent(xXITouchOwnershipEvent *in,
     return 1;
 }
 
-#define FP3232_TO_DOUBLE(x) ((double) (x).integral + (x).frac / (1ULL << 32))
-
 static int
 wireToBarrierEvent(xXIBarrierEvent *in, XGenericEventCookie *cookie)
 {
@@ -2001,6 +2105,7 @@ wireToBarrierEvent(xXIBarrierEvent *in, XGenericEventCookie *cookie)
 
     out->display    = cookie->display;
     out->type       = in->type;
+    out->serial     = cookie->serial;
     out->extension  = in->extension;
     out->evtype     = in->evtype;
     out->send_event = ((in->type & 0x80) != 0);
@@ -2020,3 +2125,91 @@ wireToBarrierEvent(xXIBarrierEvent *in, XGenericEventCookie *cookie)
 
     return 1;
 }
+
+static int
+wireToPinchEvent(xXIGesturePinchEvent *in,
+                 XGenericEventCookie *cookie)
+{
+    XIGesturePinchEvent *out;
+
+    cookie->data = out = malloc(sizeof(XIGesturePinchEvent));
+
+    out->display    = cookie->display;
+    out->type       = in->type;
+    out->serial     = cookie->serial;
+    out->extension  = in->extension;
+    out->evtype     = in->evtype;
+    out->send_event = ((in->type & 0x80) != 0);
+    out->time       = in->time;
+    out->deviceid   = in->deviceid;
+    out->sourceid   = in->sourceid;
+    out->detail     = in->detail;
+    out->root       = in->root;
+    out->event      = in->event;
+    out->child      = in->child;
+    out->root_x     = FP1616toDBL(in->root_x);
+    out->root_y     = FP1616toDBL(in->root_y);
+    out->event_x    = FP1616toDBL(in->event_x);
+    out->event_y    = FP1616toDBL(in->event_y);
+    out->delta_x    = FP1616toDBL(in->delta_x);
+    out->delta_y    = FP1616toDBL(in->delta_y);
+    out->delta_unaccel_x = FP1616toDBL(in->delta_unaccel_x);
+    out->delta_unaccel_y = FP1616toDBL(in->delta_unaccel_y);
+    out->scale      = FP1616toDBL(in->scale);
+    out->delta_angle = FP1616toDBL(in->delta_angle);
+    out->flags      = in->flags;
+
+    out->mods.base = in->mods.base_mods;
+    out->mods.locked = in->mods.locked_mods;
+    out->mods.latched = in->mods.latched_mods;
+    out->mods.effective = in->mods.effective_mods;
+    out->group.base = in->group.base_group;
+    out->group.locked = in->group.locked_group;
+    out->group.latched = in->group.latched_group;
+    out->group.effective = in->group.effective_group;
+
+    return 1;
+}
+
+static int
+wireToSwipeEvent(xXIGestureSwipeEvent *in,
+                 XGenericEventCookie *cookie)
+{
+    XIGestureSwipeEvent *out;
+
+    cookie->data = out = malloc(sizeof(XIGestureSwipeEvent));
+
+    out->display    = cookie->display;
+    out->type       = in->type;
+    out->serial     = cookie->serial;
+    out->extension  = in->extension;
+    out->evtype     = in->evtype;
+    out->send_event = ((in->type & 0x80) != 0);
+    out->time       = in->time;
+    out->deviceid   = in->deviceid;
+    out->sourceid   = in->sourceid;
+    out->detail     = in->detail;
+    out->root       = in->root;
+    out->event      = in->event;
+    out->child      = in->child;
+    out->root_x     = FP1616toDBL(in->root_x);
+    out->root_y     = FP1616toDBL(in->root_y);
+    out->event_x    = FP1616toDBL(in->event_x);
+    out->event_y    = FP1616toDBL(in->event_y);
+    out->delta_x    = FP1616toDBL(in->delta_x);
+    out->delta_y    = FP1616toDBL(in->delta_y);
+    out->delta_unaccel_x = FP1616toDBL(in->delta_unaccel_x);
+    out->delta_unaccel_y = FP1616toDBL(in->delta_unaccel_y);
+    out->flags      = in->flags;
+
+    out->mods.base = in->mods.base_mods;
+    out->mods.locked = in->mods.locked_mods;
+    out->mods.latched = in->mods.latched_mods;
+    out->mods.effective = in->mods.effective_mods;
+    out->group.base = in->group.base_group;
+    out->group.locked = in->group.locked_group;
+    out->group.latched = in->group.latched_group;
+    out->group.effective = in->group.effective_group;
+
+    return 1;
+}
